import requests
from flask import Blueprint, render_template, request, redirect
from flask import url_for, abort, Response
from iso3166 import countries
from metasrht.graphql import gql_product, gql_invoice, gql_subscription
from metasrht.graphql import gql_payment_method
from metasrht.types import PaymentStatus, PaymentCurrency
from srht.config import cfg, get_origin
from srht.crypto import encrypt_request_authorization
from srht.flask import session
from srht.graphql import exec_gql, gql_time
from srht.oauth import current_user, loginrequired
from srht.validation import Validation

billing = Blueprint('billing', __name__)
onboarding_redirect = cfg("meta.sr.ht::settings", "onboarding-redirect")

# Maximum number of invoices to display on abbreviated views
MAX_INVOICE = 4

CURRENCY_SYMBOLS = {
    PaymentCurrency.EUR: "€",
    PaymentCurrency.USD: "$",
}

CURRENCY_NAMES = {
    PaymentCurrency.EUR: "Euro",
    PaymentCurrency.USD: "US Dollar",
}

class BillingState(dict):
    def __init__(self, product_id=None):
        self.product_id = product_id

    def store(self):
        session["billing_state"] = self.__dict__

def get_address(valid):
    address = {}

    address["country"] = valid.require("country")
    valid.expect(countries.get(address["country"], None) is not None,
            "Invalid country", field="country")
    address["fullName"] = valid.optional("fullName") or None
    address["businessName"] = valid.optional("businessName") or None
    address["address1"] = valid.optional("address1") or None
    address["address2"] = valid.optional("address2") or None
    address["city"] = valid.optional("city") or None
    address["region"] = valid.optional("region") or None
    address["postcode"] = valid.optional("postcode") or None
    address["vat"] = valid.optional("vat") or None
    return address

@billing.route("/billing")
@loginrequired
def billing_GET():
    session.pop("idempotency_key", None)

    resp = exec_gql("meta.sr.ht", """
    query BillingInfo {
        me {
            subscription {
                status
                updated
                interval
                autorenew
                currency
                subtotal
                total

                payment {
                    status
                    error
                }
            }

            billingAddress {
                fullName
                businessName
                address1
                address2
                city
                region
                postcode
                country
                vat
            }

            invoices {
                results {
                    id
                    invoiceNo
                    issued
                    serviceEnd
                    total
                }
            }

            paymentMethods {
                id
                name
                created
                expires
            }

            defaultPaymentMethod { id }
        }
    }
    """)

    sub = gql_subscription(resp["me"]["subscription"])
    if not sub:
        return redirect(url_for(".billing_setup_GET"))
    invoices = [gql_invoice(i) for i in resp["me"]["invoices"]["results"]]
    if len(invoices) > MAX_INVOICE:
        invoices = invoices[:MAX_INVOICE]
    address = resp["me"]["billingAddress"]
    if address is not None and address.get("country", None) is not None:
        address["country"] = countries.get(address["country"])

    methods = resp["me"]["paymentMethods"]
    methods = [gql_payment_method(pm) for pm in methods]

    return render_template("billing.html",
            message=session.pop("message", None),
            sub=sub, invoices=invoices, address=address,
            methods=resp["me"]["paymentMethods"],
            default_method=resp["me"]["defaultPaymentMethod"],
            sym=CURRENCY_SYMBOLS[sub["currency"]])

billing_address_query = """
query BillingAddress {
    me {
        billingAddress {
            fullName
            businessName
            address1
            address2
            city
            region
            postcode
            country
            vat
        }
    }
}
"""

@billing.route("/billing/address")
@loginrequired
def billing_address_GET():
    resp = exec_gql("meta.sr.ht", billing_address_query)
    address = resp["me"]["billingAddress"]
    return render_template("billing-edit-address.html",
        address=address, countries=countries)

@billing.route("/billing/address", methods=["POST"])
@loginrequired
def billing_address_POST():
    valid = Validation(request)
    address = get_address(valid)

    if not valid.ok:
        resp = exec_gql("meta.sr.ht", billing_address_query)
        address = resp["me"]["billingAddress"]
        return render_template("billing-edit-address.html",
            address=address, countries=countries, valid=valid)

    resp = exec_gql("meta.sr.ht", """
    mutation UpdateAddress($address: BillingAddressInput!) {
        updateBillingAddress(input: $address) {
            country
        }
    }
    """, address=address, valid=valid)

    if not valid.ok:
        resp = exec_gql("meta.sr.ht", billing_address_query)
        address = resp["me"]["billingAddress"]
        return render_template("billing-edit-address.html",
            address=address, countries=countries, valid=valid)

    return redirect(url_for("billing.billing_GET"))

def billing_setup_render(valid=None):
    resp = exec_gql("meta.sr.ht", """
    query BillingSetup {
        products {
            id
            name
            prices {
                currency
                amount
            }
        }

        me {
            paymentDue
            paymentStatus

            invoices {
                results {
                    id
                    invoiceNo
                    issued
                    serviceEnd
                    total
                }
            }

            billingAddress {
                fullName
                businessName
                address1
                address2
                city
                region
                postcode
                country
                vat
            }
        }
    }
    """)

    payment_status = PaymentStatus(resp["me"]["paymentStatus"])
    if payment_status not in [
            PaymentStatus.unpaid,
            PaymentStatus.free,
            PaymentStatus.subsidized,
        ]:
        return redirect(url_for(".billing_GET"))

    payment_due = resp["me"]["paymentDue"]
    if payment_due:
        payment_due = gql_time(payment_due)

    products = [gql_product(p) for p in resp["products"]]
    invoices = [gql_invoice(i) for i in resp["me"]["invoices"]["results"]]
    if len(invoices) > MAX_INVOICE:
        invoices = invoices[:MAX_INVOICE]
    address = resp["me"]["billingAddress"]

    currencies = set()
    for product in products:
        currencies.update(PaymentCurrency(p) for p in product["prices"])

    return render_template("billing-setup.html",
        invoices=invoices, products=products,
        countries=countries, address=address,
        currencies=sorted(currencies),
        CURRENCY_NAMES=CURRENCY_NAMES,
        CURRENCY_SYMBOLS=CURRENCY_SYMBOLS,
        payment_status=payment_status,
        payment_due=payment_due,
        message=session.pop("message", None),
        **({"valid": valid} if valid else {}))

@billing.route("/billing/setup")
@loginrequired
def billing_setup_GET():
    session.pop("idempotency_key", None)
    return billing_setup_render()

@billing.route("/billing/setup", methods=["POST"])
@loginrequired
def billing_setup_POST():
    valid = Validation(request)

    address = get_address(valid)
    product_id = valid.require("product_id")
    interval = valid.require("interval")
    currency = valid.require("currency")
    with valid.catch(ValueError, "Invalid product ID", field="product_id"):
        product_id = int(product_id) if product_id else None
    if not valid.ok:
        return billing_setup_render(valid)

    idempotency_key = session.get("idempotency_key", None)

    resp = exec_gql("meta.sr.ht", """
    mutation OrderProduct(
        $productID: Int!
        $currency: Currency!,
        $address: BillingAddressInput!,
        $interval: PaymentInterval!,
        $idempotencyKey: String,
    ) {
        orderProductSubscription(
            productID: $productID,
            currency: $currency,
            interval: $interval,
            address: $address,
            autorenew: true,
            idempotencyKey: $idempotencyKey,
        ) {
            subscription {
                user {
                    paymentStatus
                }

                currency
                interval

                product {
                    id
                    name
                }

                price {
                    currency
                    amount
                }

                subtotal
            }

            billingAddress {
                fullName
                address1
                address2
                city
                region
                postcode
                country
            }

            idempotencyKey
            taxRate
            taxDue
            totalDue

            ... on StripePaymentIntent {
                publicKey
                clientSecret
            }
        }
    }
    """,
        address=address,
        productID=product_id,
        interval=interval,
        currency=currency,
        idempotencyKey=idempotency_key,
        valid=valid)

    if not valid.ok:
        return billing_setup_render(valid)

    intent = resp["orderProductSubscription"]
    sub = gql_subscription(intent["subscription"])
    product = sub["product"]
    price = sub["price"]

    session["idempotency_key"] = intent["idempotencyKey"]
    return render_template("billing-payment-intent.html",
        sub=sub, product=product, price=price, intent=intent,
        sym=CURRENCY_SYMBOLS[sub["currency"]])

@billing.route("/billing/validate-payment")
@loginrequired
def validate_payment_GET():
    intent_id = request.args.get("payment_intent")

    r = exec_gql("meta.sr.ht", """
    mutation FinalizePaymentIntent($intentID: String!) {
        finalizePaymentIntent(intentID: $intentID) {
            outcome {
                status
                error
            }
        }
    }
    """, intentID=intent_id)

    match r["finalizePaymentIntent"]["outcome"]["status"]:
        case 'SUCCEEDED' | 'PROCESSING':
            return redirect(url_for("billing.billing_complete"))
        case _:
            session["message"] = 'We were unable to complete your payment. Please try again, or contact support for assistance.'
            return redirect(url_for("billing.billing_setup_GET"))


@billing.route("/billing/complete")
@loginrequired
def billing_complete():
    return render_template("billing-complete.html",
            onboarding_redirect=onboarding_redirect)

@billing.route("/billing/new-payment-method")
@loginrequired
def new_payment_method_GET():
    resp = exec_gql("meta.sr.ht", """
    query ActiveSubscription {
        me {
            paymentStatus

            subscription {
                id
                autorenew
                currency
                interval
                subtotal

                product {
                    name
                }

                price {
                    currency
                    amount
                }
            }

            billingAddress {
                fullName
                businessName
                address1
                address2
                city
                region
                postcode
                country
                vat
            }
        }
    }
    """)
    sub = gql_subscription(resp["me"]["subscription"])
    if not sub:
        return redirect(url_for(".billing_setup_GET"))
    address = resp["me"]["billingAddress"]

    if PaymentStatus(resp["me"]["paymentStatus"]) != PaymentStatus.delinquent:
        resp = exec_gql("meta.sr.ht", """
        mutation GetSetupIntent {
            createSetupIntent {
                id

                ... on StripeSetupIntent {
                    publicKey
                    clientSecret
                }
            }
        }
        """)

        return render_template("billing-setup-intent.html",
            intent=resp["createSetupIntent"], sub=sub, address=address)
    else:
        idempotency_key = session.get("idempotency_key", None)

        # When adding a payment method to a delinquent account, we use a
        # payment intent instead of a setup intent to charge the user
        # immediately
        resp = exec_gql("meta.sr.ht", """
        mutation GetPaymentIntent(
            $subscriptionID: Int!,
            $idempotencyKey: String,
        ) {
            beginSubscriptionRenewal(
                subscriptionID: $subscriptionID,
                idempotencyKey: $idempotencyKey,
            ) {
                subscription {
                    user {
                        paymentStatus
                    }

                    currency
                    interval

                    product {
                        id
                        name
                    }

                    price {
                        currency
                        amount
                    }

                    subtotal
                }

                billingAddress {
                    fullName
                    address1
                    address2
                    city
                    region
                    postcode
                    country
                }

                idempotencyKey
                taxRate
                taxDue
                totalDue

                ... on StripePaymentIntent {
                    publicKey
                    clientSecret
                }
            }
        }
        """, subscriptionID=sub["id"], idempotencyKey=idempotency_key)

        intent = resp["beginSubscriptionRenewal"]
        sub = gql_subscription(intent["subscription"])
        product = sub["product"]
        price = sub["price"]

        session["idempotency_key"] = intent["idempotencyKey"]

        return render_template("billing-payment-intent.html",
            sub=sub, product=product, price=price, intent=intent,
            sym=CURRENCY_SYMBOLS[sub["currency"]])

@billing.route("/billing/validate-setup")
@loginrequired
def validate_setup_GET():
    intent_id = request.args.get("setup_intent")

    r = exec_gql("meta.sr.ht", """
    mutation FinalizeSetupIntent($intentID: String!) {
        finalizeSetupIntent(intentID: $intentID) {
            status
        }
    }
    """, intentID=intent_id)

    status = r["finalizeSetupIntent"]["status"]
    print(status)
    match status:
        case "SUCCEEDED":
            session["message"] = f"Your new payment method was added successfully."
        case "CANCELLED":
            session["message"] = f"Adding your new payment method was cancelled before it could be completed."

    return redirect(url_for("billing.billing_GET"))

@billing.route("/billing/remove-method/<method_id>", methods=["POST"])
@loginrequired
def remove_method_POST(method_id):
    exec_gql("meta.sr.ht", """
    mutation RemoveMethod($methodID: Int!) {
        deletePaymentMethod(id: $methodID) {
            id
        }
    }
    """, methodID=method_id)
    return redirect(url_for("billing.billing_GET"))

@billing.route("/billing/set-default-method/<method_id>", methods=["POST"])
@loginrequired
def set_default_method_POST(method_id):
    exec_gql("meta.sr.ht", """
    mutation SetDefaultMethod($methodID: Int!) {
        setDefaultPaymentMethod(id: $methodID) {
            id
        }
    }
    """, methodID=method_id)
    return redirect(url_for("billing.billing_GET"))

def change_product_render(**kwargs):
    resp = exec_gql("meta.sr.ht", """
    query BillingSetup {
        me {
            paymentStatus
            paymentDue

            subscription {
                autorenew
                currency
                status
                updated
            }
        }

        products {
            id
            name
            prices {
                currency
                amount
            }
        }
    }
    """)

    if resp["me"]["paymentStatus"] not in ["CURRENT", "DELINQUENT"]:
        return redirect(url_for(".billing_GET"))

    user = resp["me"]
    user["paymentStatus"] = PaymentStatus(user["paymentStatus"])
    user["paymentDue"] = gql_time(user["paymentDue"])
    sub = gql_subscription(user["subscription"])

    products = [gql_product(p) for p in resp["products"]]
    return render_template("billing-change-product.html",
        products=products, user=user, sub=sub, **kwargs)

@billing.route("/billing/change-product")
@loginrequired
def change_product_GET():
    return change_product_render()

@billing.route("/billing/change-product", methods=["POST"])
@loginrequired
def change_product_POST():
    valid = Validation(request)
    product_id = valid.require("product_id")
    interval = valid.require("interval")
    with valid.catch(ValueError, "Invalid product ID", field="product_id"):
        product_id = int(product_id) if product_id else None
    if not valid.ok:
        return change_product_render(valid=valid)

    resp = exec_gql("meta.sr.ht", """
    query BillingSetup {
        me {
            paymentStatus

            subscription {
                id
            }
        }
    }
    """)
    sub_id = resp["me"]["subscription"]["id"]
    status = PaymentStatus(resp["me"]["paymentStatus"])

    exec_gql("meta.sr.ht", """
    mutation ChangeProduct(
        $subID: Int!,
        $productID: Int!,
        $interval: PaymentInterval!,
    ) {
        updateBillingSubscription(
            subscriptionID: $subID,
            input: {
                productID: $productID,
                interval: $interval,
            },
        ) {
            id
        }
    }
    """, subID=sub_id, productID=product_id, interval=interval)

    if status == PaymentStatus.delinquent:
        return redirect(url_for(".new_payment_method_GET"))

    return redirect(url_for(".billing_GET"))

@billing.route("/billing/cancel")
@loginrequired
def cancel_GET():
    resp = exec_gql("meta.sr.ht", """
    query BillingSetup {
        me {
            paymentStatus
            paymentDue

            subscription {
                status
                updated
            }
        }
    }
    """)
    user = resp["me"]
    user["paymentStatus"] = PaymentStatus(user["paymentStatus"])
    user["paymentDue"] = gql_time(user["paymentDue"])
    user["subscription"] = gql_subscription(user["subscription"])

    if user["paymentStatus"] not in [PaymentStatus.current, PaymentStatus.delinquent]:
        return redirect(url_for(".billing_GET"))

    return render_template("billing-cancel.html",
               user=user, sub=user["subscription"])

@billing.route("/billing/cancel", methods=["POST"])
@loginrequired
def cancel_POST():
    resp = exec_gql("meta.sr.ht", """
    query BillingSetup {
        me {
            subscription {
                id
            }
        }
    }
    """)
    sub_id = resp["me"]["subscription"]["id"]

    exec_gql("meta.sr.ht", """
    mutation CancelBilling($subscriptionID: Int!) {
        cancelBillingSubscription(subscriptionID: $subscriptionID) {
            id
        }
    }
    """, subscriptionID=sub_id)

    return redirect(url_for("billing.cancelled_GET"))

@billing.route("/billing/cancelled")
@loginrequired
def cancelled_GET():
    resp = exec_gql("meta.sr.ht", """
    query BillingSetup {
        me {
            paymentStatus
            paymentDue
        }
    }
    """)
    user = resp["me"]
    user["paymentStatus"] = PaymentStatus(user["paymentStatus"])
    user["paymentDue"] = gql_time(user["paymentDue"])
    return render_template("billing-cancelled.html", user=user)

@billing.route("/billing/invoices")
@loginrequired
def invoices_GET():
    cursor = request.args.get("next")

    resp = exec_gql("meta.sr.ht", """
    query Invoices($cursor: Cursor) {
        me {
            invoices(cursor: $cursor) {
                cursor
                results {
                    id
                    invoiceNo
                    issued
                    total
                    serviceEnd
                }
            }
        }
    }
    """, cursor=cursor)

    invoices = [gql_invoice(i) for i in resp["me"]["invoices"]["results"]]
    cursor = resp["me"]["invoices"]["cursor"]

    return render_template("billing-invoices.html",
        invoices=invoices, cursor=cursor)

@billing.route("/billing/invoice/<int:invoice_id>", methods=["POST"])
@loginrequired
def invoice_POST(invoice_id):
    origin = cfg("meta.sr.ht", "api-origin", default=get_origin("meta.sr.ht"))
    headers = {
        "X-Forwarded-For": ", ".join(request.access_route),
        **encrypt_request_authorization(user=current_user),
    }
    r = requests.post(f"{origin}/query/invoice/{invoice_id}",
            headers=headers, data=request.form)
    headers = [('Content-Disposition', r.headers["Content-Disposition"])]
    return Response(r.content, mimetype="application/pdf", headers=headers)
